#include <stdlib.h>
#include <string.h>
#include "functions.h"
#include "gldecoder.h"
#include "packet.h"

static int decode_func(GlDecoder *, unsigned short, GlPacket *);
static int decode_gldfunc(GlDecoder *, unsigned short, GlPacket *);

GlDecoder *gldecoder_new(void *user_data, void (*destroy)(void *))
{
	GlDecoder *dec;

	dec = (GlDecoder *)malloc(sizeof(GlDecoder));
	memset(dec, 0, sizeof(GlDecoder));
	dec->user_data = user_data;
	dec->destroy = destroy;

	return dec;
}

void gldecoder_delete(GlDecoder *dec)
{
	if(dec->destroy)
		dec->destroy(dec->user_data);
	free(dec);
}

int gldecoder_decode(GlDecoder *dec, const char *data, unsigned len)
{
	GlPacket *pkt;
	unsigned short func;

	pkt = packet_receive_str(data, &len);
	if(!pkt)
		return -1;

	packet_read_short(pkt, (short *)&func);

	int ret = 0;

	if(func&0x8000)
		ret = decode_gldfunc(dec, func, pkt);
	else
		ret = decode_func(dec, func, pkt);
	if(ret<0)
		return -1;

	return len;
}

typedef const void *pointer;
typedef const char *string;

#include "gensrc/gldecoder.funcs"

static void decode_gldError(GlDecoder *dec, GlPacket *pkt)
{
	GLenum code;
	packet_read_int(pkt, (int *)&code);
	if(dec->gldError)
		dec->gldError(dec->user_data, code);
	else if(dec->unhandled)
		dec->unhandled(dec->user_data, FUNC_GLDERROR);
}

static void decode_gldBreak(GlDecoder *dec, GlPacket *pkt)
{
	unsigned short func;
	unsigned char flag;
	packet_read_short(pkt, (short *)&func);
	packet_read_char(pkt, (char *)&flag);
	if(dec->gldBreak)
		dec->gldBreak(dec->user_data, func, flag);
	else if(dec->unhandled)
		dec->unhandled(dec->user_data, FUNC_GLDBREAK);
}

static int decode_gldfunc(GlDecoder *dec, unsigned short func, GlPacket *pkt)
{
	switch(func)
	{
	case FUNC_GLDERROR: decode_gldError(dec, pkt); break;
	case FUNC_GLDBREAK: decode_gldBreak(dec, pkt); break;
	default: return -1;
	}
	return 0;
}
